from __future__ import absolute_import
from __future__ import unicode_literals

import getopt
import hashlib
import logging
import os
import sys

from swjsq._compat import text_type
from swjsq.core import SWJSQClient, fast_d1ck
from swjsq.core import TYPE_NORMAL_ACCOUNT, TYPE_NUM_ACCOUNT
from swjsq.exceptions import APIError, LoginError, UpgradeError


class NoCredentialsError(RuntimeError):
    pass


class Arguments(object):
    def __init__(self):
        self.account_file_encrypted = '.swjsq.account'
        self.account_file_plain = 'swjsq.account.txt'


def show_usage():
    options = [
        ('-h, --help', 'show this help message and exit'),
    ]

    print('usage: {0} [OPTIONS]'.format(sys.argv[0]))
    print()
    print('options:')
    for opt, description in options:
        print('  {0}\t{1}'.format(opt, description))


def parse_args():
    try:
        long_opts = ['help']
        opts, args = getopt.getopt(sys.argv[1:], 'h', long_opts)
    except getopt.GetoptError as err:
        print(err)
        show_usage()
        sys.exit(2)

    args = Arguments()

    for o, a in opts:
        if o in ('-h', '--help'):
            show_usage()
            sys.exit()
        else:
            assert False, 'Unhandled option'

    return args


def setup_logging():
    fh = logging.FileHandler('swjsq.log')
    fh.setLevel(logging.DEBUG)

    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)

    FMT = '%(asctime)s %(levelname)s %(message)s'
    DATE_FMT = '%Y-%m-%d %H:%M:%S'
    formatter = logging.Formatter(FMT, DATE_FMT)
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)

    logger = logging.getLogger()

    logger.addHandler(fh)
    logger.addHandler(ch)

    logger.setLevel(logging.DEBUG)


class Credentials(object):
    def __init__(self, login_type, uid, password_hash):
        self.login_type = login_type
        self.uid = uid
        self.password_hash = password_hash


def load_credentials_from_file(path, skip_password_hash=False):
    with open(path, 'rb') as f:
        uid, pwd = f.read().strip().split(b',')
    if not skip_password_hash:
        pwd = hashlib.md5(pwd).hexdigest()
    # return type of hexdigest is different between PY2 and PY3
    if isinstance(pwd, text_type):
        pwd = pwd.encode('utf-8')
    return uid, pwd


def load_credentials_from_env():
    uid, pwd = os.getenv('XUNLEI_UID'), os.getenv('XUNLEI_PASSWD')
    if not uid or not pwd:
        raise RuntimeError('Environment variables not set')
    # type of environment variable is different between PY2 and PY3
    if isinstance(uid, text_type):
        uid = uid.encode(sys.getfilesystemencoding())
    if isinstance(pwd, text_type):
        pwd = pwd.encode(sys.getfilesystemencoding())
    pwd = hashlib.md5(pwd).hexdigest()
    # return type of hexdigest is different between PY2 and PY3
    if isinstance(pwd, text_type):
        pwd = pwd.encode('utf-8')
    return uid, pwd


def load_credentials(account_file_plain, account_file_encrypted):
    '''Try to load credentials.

    :param account_file_plain: Path to plain text credentials
    :param account_file_encrypted: Path to encrypted credentials
    :returns: the Credentials object loaded
    :raises NoCredentialsError: No credentials can be loaded.
    '''
    try:
        uid, pwd_md5 = load_credentials_from_file(account_file_plain)
        return Credentials(TYPE_NORMAL_ACCOUNT, uid, pwd_md5)
    except Exception as e:
        logging.debug('load_credentials_from_file plain: %s', e)

    try:
        uid, pwd_md5 = load_credentials_from_file(account_file_encrypted,
                                                  skip_password_hash=True)
        return Credentials(TYPE_NORMAL_ACCOUNT, uid, pwd_md5)
    except Exception as e:
        logging.debug('load_credentials_from_file encrypted: %s', e)

    try:
        uid, pwd_md5 = load_credentials_from_env()
        return Credentials(TYPE_NORMAL_ACCOUNT, uid, pwd_md5)
    except Exception as e:
        logging.debug('load_credentials_from_env: %s', e)
        raise NoCredentialsError()


def save_credentials(account_file_encrypted, credentials):
    content = '{0},{1}'.format(credentials.uid,
                               credentials.password_hash)
    with open(account_file_encrypted, 'w') as f:
        f.write(content)


def main():
    try:
        # Arguments
        args = parse_args()

        # Logging
        setup_logging()

        # Login
        credentials = load_credentials(args.account_file_plain,
                                       args.account_file_encrypted)

        client = SWJSQClient()
        client.login(credentials.uid, credentials.password_hash,
                     credentials.login_type)
        logging.info('Login xunlei succeeded.')

        # Save encrypted credentials
        if credentials.login_type != TYPE_NUM_ACCOUNT:
            try:
                os.remove(args.account_file_plain)
            except Exception:
                pass
            save_credentials(args.account_file_encrypted, credentials)

        # Routine
        fast_d1ck(client, credentials.password_hash)
    except NoCredentialsError:
        logging.error('No credentials provided.')
    except LoginError as e:
        logging.error('Login Error: (%d) %s',
                      e.errno, e.message or 'Unknown')
    except APIError as e:
        logging.error('API Error %s: (%d) %s',
                      e.command, e.errno, e.message or 'Unknown')
    except UpgradeError as e:
        logging.error('Upgrade Error: %s', e.message)
    except KeyboardInterrupt:
        logging.info('Stopping')


if __name__ == '__main__':
    main()



from __future__ import absolute_import
from __future__ import unicode_literals

import logging
import ssl
import sys
import time

from swjsq._compat import text_type
from swjsq._compat import iteritems, range
from swjsq._compat import parse, request
from swjsq._compat import URLError


logger = logging.getLogger(__name__)


def get(url, params=None, body=None, headers=None, max_tries=3):
    '''Get result of HTTP request
    :param url: URL of the target
    :param params: optional query string as dict
    :param body: optional request body as binary type or ascii text
    :param headers: optional request headers as dict
    :param max_tries: total count of failed tries before raising error
    :returns: body of the response as binary type
    :raises URLError: request failed even after retries
    '''
    if params:
        query_string = parse.urlencode(params)
        delimiter = '&' if '?' in url else '?'
        url = '{0}{1}{2}'.format(url, delimiter, query_string)

    req = request.Request(url)
    for k, v in iteritems(headers or {}):
        req.add_header(k, v)
    if isinstance(body, text_type):
        body = body.encode('ascii')

    # Xunlei uses a self-signed certificate
    # which would be rejected by default in Python 2.7.9+
    extra_options = {}
    if sys.version_info >= (2, 7, 9):
        context = ssl.create_default_context()
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        extra_options['context'] = context

    sleep_increment = 2
    for i in range(1, max_tries + 1):
        try:
            resp = request.urlopen(req, data=body, **extra_options)
        except URLError as e:
            logger.debug('#%d request failed: %s', i, e)
            if i == max_tries:
                raise
            time.sleep(i * sleep_increment)
        else:
            break

    return resp.read()



from __future__ import absolute_import
from __future__ import unicode_literals

import binascii
import functools
import logging

from swjsq._compat import long, binary_type, iterbytes


logger = logging.getLogger(__name__)


def _cache_recent(f):
    '''Cache the most recent argument/result'''
    data = {
        'args': None,
        'result': None,
    }

    @functools.wraps(f)
    def wrapped(*args):
        if data.get('args') != args:
            data['args'] = args
            data['result'] = f(*args)
        return data['result']
    return wrapped


@_cache_recent
def rsa_encrypt(pubexp, mod, payload):
    '''RSA encrypt
    :param pubexp: RSA pubexp as hex string
    :param mod: RSA mod as hex string
    :param payload: binary data to be encrypted
    :returns: cipher as hex string
    :raises TypeError: if payload is not binary type
    '''
    if not isinstance(payload, binary_type):
        raise TypeError('payload should be of binary type')

    pubexp = long(pubexp, 16)
    mod = long(mod, 16)

    try:
        return _rsa_encrypt_pgcrypto(pubexp, mod, payload)
    except ImportError:
        logger.debug('pycrypto not found, using manual rsa')
        return _rsa_encrypt_python(pubexp, mod, payload)


def _rsa_encrypt_pgcrypto(pubexp, mod, payload):
    from Crypto.PublicKey import RSA

    rsa = RSA.construct((mod, pubexp))
    cipher = rsa.encrypt(payload, None)[0]

    return binascii.hexlify(cipher).decode('ascii')


# https://github.com/mengskysama/XunLeiCrystalMinesMakeDie/blob/master/run.py
def _rsa_encrypt_python(pubexp, mod, payload):
    def modpow(b, e, m):
        result = 1
        while (e > 0):
            if e & 1:
                result = (result * b) % m
            e = e >> 1
            b = (b * b) % m
        return result

    def binary_to_int(binary):
        str_int = 0
        for c in iterbytes(binary):
            str_int = str_int << 8
            str_int += c
        return str_int

    cipher = modpow(binary_to_int(payload), pubexp, mod)
    return '{0:0256X}'.format(cipher)



from __future__ import absolute_import

import collections
import logging
import json
import time
import hashlib
import atexit

from swjsq._compat import binary_type, text_type
from swjsq._compat import URLError
from swjsq.exceptions import APIError, LoginError, SWJSQError, UpgradeError
from swjsq.http import get as http_get
from swjsq.rsa import rsa_encrypt
from swjsq.utils import get_mac

logger = logging.getLogger(__name__)


RSA_MOD = u'AC69F5CCC8BDE47CD3D371603748378C9CFAD2938A6B021E0E191013975AD683F5CBF9ADE8BD7D46B4D2EC2D78AF146F1DD2D50DC51446BB8880B8CE88D476694DFC60594393BEEFAA16F5DBCEBE22F89D640F5336E42F587DC4AFEDEFEAC36CF007009CCCE5C1ACB4FF06FBA69802A8085C2C54BADD0597FC83E6870F1E36FD'
RSA_PUBEXP = u'010001'

PROTOCOL_VERSION = 108
FALLBACK_MAC = '00:00:00:00:00:00'
FALLBACK_INTERFACE = u'119.147.41.210:80'

TYPE_NORMAL_ACCOUNT = 0
TYPE_NUM_ACCOUNT = 1


class Session(object):
    def __init__(self, login_response):
        self.raw = login_response

    @property
    def user_id(self):
        return self.raw.get(u'userID')

    @property
    def session_id(self):
        return self.raw.get(u'sessionID')

    @property
    def can_upgrade(self):
        if self.raw.get(u'isVip'):
            return True
        if self.raw.get(u'payId') in [5, 702]:
            return True
        return False

    @property
    def is_subaccount(self):
        return self.raw.get('isSubAccount', False)


class Bandwidth(object):
    BandwidthDetail = collections.namedtuple(u'BandwidthDetail',
                                             u'upstream downstream')

    def __init__(self, response):
        self.raw = response

    @property
    def can_upgrade(self):
        return self.raw.get(u'can_upgrade', False)

    @property
    def dial_account(self):
        return self.raw.get(u'dial_account')

    @property
    def original(self):
        detail = self.raw[u'bandwidth']  # FIXME: what if no such field?
        return self.BandwidthDetail(detail[u'upstream'], detail[u'downstream'])

    @property
    def max(self):
        detail = self.raw[u'max_bandwidth']  # FIXME: what if no such field?
        return self.BandwidthDetail(detail[u'upstream'], detail[u'downstream'])

    @property
    def province_name(self):
        return self.raw.get(u'province_name')

    @property
    def sp_name(self):
        return self.raw.get(u'sp_name')


class XunleiClient(object):
    def __init__(self, business_type, client_version, rsa_key=None):
        self.session = None
        self.business_type = business_type
        self.client_version = client_version
        self.rsa_pubexp, self.rsa_mod = rsa_key or (RSA_PUBEXP, RSA_MOD)

    @property
    def peer_id(self):
        peer_id = getattr(self, '_peer_id', None)
        if peer_id is None:
            mac = get_mac() or FALLBACK_MAC
            peer_id = mac.replace(':', '').upper() + '004V'
            setattr(self, '_peer_id', peer_id)
        return peer_id

    def _json_http_req(self, url, params=None, body=None, headers=None,
                       max_tries=3, encoding=None):
        encoding = encoding or u'utf-8'
        response = http_get(url, params, body, headers, max_tries)
        return json.loads(response.decode(encoding))

    def _xunlei_request(self, payload):
        headers = {
            u'Content-Type': u'',
            u'Connection': u'Keep-Alive',
            u'Accept-Encoding': u'gzip',
            u'User-Agent': (
                u'android-async-http/xl-acc-sdk/version-1.6.1.177600'
            ),
        }
        url = u'https://login.mobile.reg2t.sandai.net:443/'
        return self._json_http_req(url, body=payload,
                                   headers=headers, encoding=u'gbk')

    def login(self, uname, pwd_md5, login_type=TYPE_NORMAL_ACCOUNT,
              verify_key=None, verify_code=None):
        verify_key = verify_key or u''
        verify_code = verify_code or u''

        # just generate a 32-bit string
        fake_device_id = hashlib.md5(("%s23333" % pwd_md5).encode('utf-8')).hexdigest()

        # sign = div.10?.device_id + md5(sha1(packageName + businessType + md5(a protocolVersion specific GUID)))
        device_sign = "div100.%s%s" % (fake_device_id, hashlib.md5(
            hashlib.sha1(
                ("%scom.xunlei.vip.swjsq68700d1872b772946a6940e4b51827e8af" % fake_device_id).encode('utf-8')
            ).hexdigest().encode('utf-8')
        ).hexdigest())

        if isinstance(uname, binary_type):
            username = uname.decode('utf-8')
        elif not isinstance(uname, text_type):
            username = text_type(uname)
        else:
            username = uname

        payload = json.dumps({
            u'protocolVersion': PROTOCOL_VERSION,  # 109
            u'sequenceNo': 1000001,
            u'platformVersion': 1,
            u'sdkVersion': 177550,  # 177600
            u'peerID': self.peer_id,
            u'businessType': self.business_type,
            u'clientVersion': self.client_version,
            u'devicesign': device_sign,
            u'isCompressed': 0,
            u'cmdID': 1,
            u'userName': username,
            u'passWord': rsa_encrypt(self.rsa_pubexp, self.rsa_mod, pwd_md5),
            u'loginType': login_type,
            u'sessionID': u'',
            u'verifyKey': verify_key,
            u'verifyCode': verify_code,
            u'appName': u'ANDROID-com.xunlei.vip.swjsq',
            u'rsaKey': {
                u'e': self.rsa_pubexp,
                u'n': self.rsa_mod,
            },
            u'extensionList': u'',
        })

        # TODO: Verification code handling
        # If "errorCode" is 6, "errorDescUrl" contains URL of the verification
        # code image. Access the URL and we can get the image, and the
        # "VERIFY_KEY" from cookie. Next time we send the login request, fill
        # in the "verifyKey" and "verifyCode".
        response = self._xunlei_request(payload)

        code = response.get(u'errorCode')
        if code != 0:
            message = response.get(u'errorDesc')
            logger.debug(u'Login failed: (%d) %s', code, message or 'Unknown')
            raise LoginError(code, message)

        logger.debug(u'isVip:%s payId:%s payName:%s',
                     response.get(u'isVip'),
                     response.get(u'payId'), response.get(u'payName'))

        self.session = Session(response)

        return self.session

    def renew(self):
        payload = json.dumps({
            u'protocolVersion': PROTOCOL_VERSION,
            u'sequenceNo': 1000000,
            u'platformVersion': 1,
            u'peerID': self.peer_id,
            u'businessType': self.business_type,
            u'clientVersion': self.client_version,
            u'isCompressed': 0,
            u'cmdID': 11,
            u'userID': self.session.user_id,
            u'sessionID': self.session.session_id,
        })
        response = self._xunlei_request(payload)

        code = response.get(u'errorCode')
        if code != 0:
            message = response.get(u'errorDesc')
            logger.debug(u'Renew failed: (%d) %s', code, message or u'Unknown')
            raise LoginError(code, message)

        return response


class SWJSQClient(XunleiClient):
    def __init__(self):
        business_type = 68  # Constant. Probably for SWJSQ
        client_version = "2.0.3.4"
        super(SWJSQClient, self).__init__(business_type, client_version)

        self.api_url = self._get_api_url()
        logger.debug(u'API URL: %s', self.api_url)

    def _get_api_url(self):
        # Try to connect to the first available query portal server
        PORTAL_URLS = [
            u'http://api.portal.swjsq.vip.xunlei.com:81/v2/queryportal',
            u'http://api2.portal.swjsq.vip.xunlei.com:81/v2/queryportal',
        ]
        for url in PORTAL_URLS:
            try:
                portal = self._json_http_req(url, max_tries=1)
            except URLError:
                pass
            else:
                break
        else:
            logger.warn(u'queryportal failed: no portal server available')
            return FALLBACK_INTERFACE

        errno = portal.get(u'errno')
        if errno != 0:
            message = portal.get(u'message', u'Unknown')
            logger.warn(u'queryportal failed: (%d) %s', errno, message)
            return FALLBACK_INTERFACE

        try:
            ip = portal[u'interface_ip']
            port = portal[u'interface_port']
        except KeyError as e:
            logger.warn(u'queryportal format error: %s', e)
            return FALLBACK_INTERFACE

        return u'{0}:{1}'.format(ip, port)

    def _execute(self, cmd, extras=None):
        # for 'bandwidth' command, `userid` and `sessionid` are not mandatory
        params = {
            u'client_type': u'android-swjsq-{0}'.format(self.client_version),
            u'peerid': self.peer_id,
            u'time_and': time.time() * 1000,
            u'client_version': u'androidswjsq-{0}'.format(self.client_version),
            u'userid': self.session.user_id,
            u'sessionid': self.session.session_id,
            u'os': u'android-5.0.1.23SmallRice',
        }
        if extras:
            params.update(extras)

        url = u'http://{0}/v2/{1}'.format(self.api_url, cmd)
        headers = {
            u'Content-Type': u'',
            u'Connection': u'Keep-Alive',
            u'Accept-Encoding': u'gzip',
            u'User-Agent': (
                u'Dalvik/2.1.0 '
                u'(Linux; U; Android 5.0.1; SmallRice Build/LRX22C)'
            ),
        }
        response = self._json_http_req(url, params=params, headers=headers)

        errno = response.get('errno')
        if errno:
            message = response.get('richmessage')
            if not message:
                message = response.get('message')
            raise APIError(cmd, errno, message)

        return response

    def get_bandwidth(self):
        response = self._execute(u'bandwidth')
        bandwidth = Bandwidth(response)

        KB_PER_MB = 1024
        logger.info(u'To Upgrade: %s%s Down %dM -> %dM, Up %dM -> %dM',
                    bandwidth.province_name, bandwidth.sp_name,
                    bandwidth.original.downstream / KB_PER_MB,
                    bandwidth.max.downstream / KB_PER_MB,
                    bandwidth.original.upstream / KB_PER_MB,
                    bandwidth.max.upstream / KB_PER_MB)
        return bandwidth

    def upgrade(self, bandwidth):
        extras = {
            u'user_type': 1,
            u'dial_account': bandwidth.dial_account,
        }
        response = self._execute(u'upgrade', extras=extras)
        logger.info(u'Upgrade done: Down %dM, Up %dM',
                    response[u'bandwidth'][u'downstream'],
                    response[u'bandwidth'][u'upstream'])
        return response

    def recover(self, bandwidth):
        extras = {
            u'dial_account': bandwidth.dial_account,
        }
        return self._execute(u'recover', extras=extras)

    def heartbeat(self):
        return self._execute(u'keepalive')


def fast_d1ck(client, password_hash):
    if client.session.is_subaccount:
        raise UpgradeError(u'Subaccount cannot upgrade')

    if not client.session.can_upgrade:
        logger.warn(u'You are probably not Xunlei VIP')

    bandwidth = client.get_bandwidth()
    if not bandwidth.can_upgrade:
        logger.error(u'Does not support upgrading.')
        raise UpgradeError(u'Bandwidth cannot upgrade')

    def _atexit_func():
        logger.info(u'Sending recover request')
        try:
            client.recover(bandwidth)
        except KeyboardInterrupt:
            logger.info(u'Secondary ctrl+c pressed, exiting')
        else:
            logger.info(u'Recover done. exiting')

    atexit.register(_atexit_func)
    i = 0
    while True:
        try:
            # i=1~17 keepalive, renew session, i++
            # i=18 (3h) re-upgrade, i:=0
            # i=100 login, i:=36
            if i == 100:
                try:
                    client.login(client.session.user_id,
                                 password_hash,
                                 TYPE_NUM_ACCOUNT)
                except SWJSQError:
                    logger.error('login_xunlei failed')
                i = 18

            if i % 18 == 0:  # 3h
                logger.info('Initializing upgrade')
                if i:  # not first time
                    client.recover(bandwidth)
                    time.sleep(5)
                _ = client.upgrade(bandwidth)
                i = 0
            else:
                try:
                    client.renew()
                except SWJSQError:
                    logger.error('renew_xunlei failed')
                    i = 100
                    continue
                _ = client.heartbeat()

            logger.debug('%s', _)
        except APIError as e:
            logger.error('APIError %s: (%d) %s', e.command, e.errno, e.message)
            if e.errno == 513:  # not exist channel: re-upgrade
                i = 100
                continue
            elif e.errno in (717, 718):  # account auth session failed
                i = 100  # The App also asks to relogin
                continue
            elif e.errno == 812:
                logger.info('Already upgraded, continuing')
                i = 0
            else:
                time.sleep(300)
        except Exception:
            logger.exception('Unexpected')

        i += 1
        time.sleep(600)  # 10 min



class SWJSQError(RuntimeError):
    pass


class APIError(SWJSQError):
    def __init__(self, command, errno, message):
        self.command = command
        self.errno = errno
        self.message = message


class LoginError(SWJSQError):
    def __init__(self, errno, message):
        self.errno = errno
        self.message = message


class UpgradeError(SWJSQError):
    def __init__(self, message):
        self.message = message



# vim: set fileencoding=utf-8
from setuptools import setup

from swjsq import __version__


with open('README.rst') as f:
    long_description = f.read()


url = 'https://github.com/timothyqiu/python-swjsq'
download_url = '{0}/archive/{1}.tar.gz'.format(url, __version__)

setup(
    name='swjsq',
    version=__version__,
    packages=['swjsq'],
    entry_points={
        'console_scripts': [
            'swjsq = swjsq.cli:main',
        ],
    },
    license='MIT',
    description='A command line client for Xunlei SWJSQ (aka. 迅雷快鸟)',
    long_description=long_description,
    author='Timothy Qiu',
    author_email='timothyqiu32@gmail.com',
    url=url,
    download_url=download_url,
    keywords=['xunlei'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Internet',
        'Topic :: Utilities',
    ],
)



from __future__ import absolute_import
from __future__ import unicode_literals

import os
import re


def get_mac(nic=None):
    '''Gets MAC address of specific device.
    :param nic: device name as string, or None if any device is ok
    :returns: MAC address as colon separated string, or None on failure
    '''
    if os.name == 'nt':
        cmd = 'ipconfig /all'
        splt = '-'
    elif os.name == "posix":
        if os.path.exists('/usr/bin/ip') or os.path.exists('/bin/ip'):
            if nic:
                cmd = 'ip link show dev {0}'.format(nic)
            else:
                # Unfortunately, loopback interface always comes first
                # So we have to grep it out
                cmd = 'ip link show up | grep -v loopback'
        else:
            cmd = 'ifconfig {0}'.format(nic or '-a')
        splt = ':'
    else:
        return None

    try:
        output = os.popen(cmd).read()
        if not output:
            return None

        raw_pattern = r'((?:[0-9A-Fa-f]{{2}}{splt}){{5}}[0-9A-Fa-f]{{2}})'
        pattern = raw_pattern.format(splt=splt)
        matches = re.findall(pattern, output)
        if not matches:
            return None

        return matches[0].replace(splt, ':')
    except Exception:
        pass

    return None



from __future__ import absolute_import

import functools
import itertools
import sys


PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

if PY2:
    import __builtin__

    text_type = __builtin__.unicode
    binary_type = str

    iterbytes = functools.partial(itertools.imap, ord)

    def iteritems(d, **kwargs):
        return d.iteritems(**kwargs)

    range = __builtin__.xrange
    long = __builtin__.long

    import urllib
    import urllib2
    parse = urllib
    request = urllib2
    URLError = urllib2.URLError
else:
    import builtins

    text_type = str
    binary_type = bytes

    iterbytes = iter

    def iteritems(d, **kwargs):
        return d.items(**kwargs)

    range = builtins.range
    long = builtins.int

    import urllib.error
    import urllib.parse
    import urllib.request
    parse = urllib.parse
    request = urllib.request
    URLError = urllib.error.URLError
